@@ -9,6 +9,7 @@
  * %End-Header%
  */
 
+#include "config.h"
 #include <stdio.h>
 #include <string.h>
 #if HAVE_UNISTD_H
@@ -34,14 +35,17 @@ errcode_t ext2fs_check_desc(ext2_filsys fs)
 	ext2fs_block_bitmap bmap;
 	errcode_t retval;
 	dgrp_t i;
-	blk_t first_block = fs->super->s_first_data_block;
-	blk_t last_block = fs->super->s_blocks_count-1;
-	blk_t blk, b;
-	int j;
+	blk64_t first_block = fs->super->s_first_data_block;
+	blk64_t last_block = ext2fs_blocks_count(fs->super)-1;
+	blk64_t blk, b;
+	unsigned int j;
 
 	EXT2_CHECK_MAGIC(fs, EXT2_ET_MAGIC_EXT2FS_FILSYS);
 
-	retval = ext2fs_allocate_block_bitmap(fs, "check_desc map", &bmap);
+	if (EXT2_DESC_SIZE(fs->super) & (EXT2_DESC_SIZE(fs->super) - 1))
+		return EXT2_ET_BAD_DESC_SIZE;
+
+	retval = ext2fs_allocate_subcluster_bitmap(fs, "check_desc map", &bmap);
 	if (retval)
 		return retval;
 
@@ -51,38 +55,36 @@ errcode_t ext2fs_check_desc(ext2_filsys fs)
 	for (i = 0; i < fs->group_desc_count; i++) {
 		if (!EXT2_HAS_INCOMPAT_FEATURE(fs->super,
 					       EXT4_FEATURE_INCOMPAT_FLEX_BG)) {
-			first_block = ext2fs_group_first_block(fs, i);
-			last_block = ext2fs_group_last_block(fs, i);
-			if (i == (fs->group_desc_count - 1))
-				last_block = fs->super->s_blocks_count-1;
+			first_block = ext2fs_group_first_block2(fs, i);
+			last_block = ext2fs_group_last_block2(fs, i);
 		}
 
 		/*
 		 * Check to make sure the block bitmap for group is sane
 		 */
-		blk = fs->group_desc[i].bg_block_bitmap;
+		blk = ext2fs_block_bitmap_loc(fs, i);
 		if (blk < first_block || blk > last_block ||
-		    ext2fs_test_block_bitmap(bmap, blk)) {
+		    ext2fs_test_block_bitmap2(bmap, blk)) {
 			retval = EXT2_ET_GDESC_BAD_BLOCK_MAP;
 			goto errout;
 		}
-		ext2fs_mark_block_bitmap(bmap, blk);
+		ext2fs_mark_block_bitmap2(bmap, blk);
 
 		/*
 		 * Check to make sure the inode bitmap for group is sane
 		 */
-		blk = fs->group_desc[i].bg_inode_bitmap;
+		blk = ext2fs_inode_bitmap_loc(fs, i);
 		if (blk < first_block || blk > last_block ||
-		    ext2fs_test_block_bitmap(bmap, blk)) {
+		    ext2fs_test_block_bitmap2(bmap, blk)) {
 			retval = EXT2_ET_GDESC_BAD_INODE_MAP;
 			goto errout;
 		}
-		ext2fs_mark_block_bitmap(bmap, blk);
+		ext2fs_mark_block_bitmap2(bmap, blk);
 
 		/*
 		 * Check to make sure the inode table for group is sane
 		 */
-		blk = fs->group_desc[i].bg_inode_table;
+		blk = ext2fs_inode_table_loc(fs, i);
 		if (blk < first_block ||
 		    ((blk + fs->inode_blocks_per_group - 1) > last_block)) {
 			retval = EXT2_ET_GDESC_BAD_INODE_TABLE;
@@ -90,11 +92,11 @@ errcode_t ext2fs_check_desc(ext2_filsys fs)
 		}
 		for (j = 0, b = blk; j < fs->inode_blocks_per_group;
 		     j++, b++) {
-			if (ext2fs_test_block_bitmap(bmap, b)) {
+			if (ext2fs_test_block_bitmap2(bmap, b)) {
 				retval = EXT2_ET_GDESC_BAD_INODE_TABLE;
 				goto errout;
 			}
-			ext2fs_mark_block_bitmap(bmap, b);
+			ext2fs_mark_block_bitmap2(bmap, b);
 		}
 	}
 errout:
