@@ -5,14 +5,15 @@
  * Copyright (C) 2000 Theodore Ts'o.
  *
  * Note: this uses the POSIX IO interfaces, unlike most of the other
- * functions in this library.  So sue me.  
+ * functions in this library.  So sue me.
  *
  * %Begin-Header%
- * This file may be redistributed under the terms of the GNU Public
- * License.
+ * This file may be redistributed under the terms of the GNU Library
+ * General Public License, version 2.
  * %End-Header%
  */
 
+#include "config.h"
 #include <stdio.h>
 #include <string.h>
 #if HAVE_UNISTD_H
@@ -62,16 +63,17 @@ errcode_t ext2fs_image_inode_write(ext2_filsys fs, int fd, int flags)
 {
 	unsigned int	group, left, c, d;
 	char		*buf, *cp;
-	blk_t		blk;
+	blk64_t		blk;
 	ssize_t		actual;
 	errcode_t	retval;
+	off_t		r;
 
 	buf = malloc(fs->blocksize * BUF_BLOCKS);
 	if (!buf)
 		return ENOMEM;
-	
+
 	for (group = 0; group < fs->group_desc_count; group++) {
-		blk = fs->group_desc[(unsigned)group].bg_inode_table;
+		blk = ext2fs_inode_table_loc(fs, (unsigned)group);
 		if (!blk) {
 			retval = EXT2_ET_MISSING_INODE_TABLE;
 			goto errout;
@@ -81,7 +83,7 @@ errcode_t ext2fs_image_inode_write(ext2_filsys fs, int fd, int flags)
 			c = BUF_BLOCKS;
 			if (c > left)
 				c = left;
-			retval = io_channel_read_blk(fs->io, blk, c, buf);
+			retval = io_channel_read_blk64(fs->io, blk, c, buf);
 			if (retval)
 				goto errout;
 			cp = buf;
@@ -96,7 +98,11 @@ errcode_t ext2fs_image_inode_write(ext2_filsys fs, int fd, int flags)
 					blk++;
 					left--;
 					cp += fs->blocksize;
-					lseek(fd, fs->blocksize, SEEK_CUR);
+					r = lseek(fd, fs->blocksize, SEEK_CUR);
+					if (r < 0) {
+						retval = errno;
+						goto errout;
+					}
 					continue;
 				}
 				/* Find non-zero blocks */
@@ -131,21 +137,21 @@ errout:
 /*
  * Read in the inode table and stuff it into place
  */
-errcode_t ext2fs_image_inode_read(ext2_filsys fs, int fd, 
+errcode_t ext2fs_image_inode_read(ext2_filsys fs, int fd,
 				  int flags EXT2FS_ATTR((unused)))
 {
 	unsigned int	group, c, left;
 	char		*buf;
-	blk_t		blk;
+	blk64_t		blk;
 	ssize_t		actual;
 	errcode_t	retval;
 
 	buf = malloc(fs->blocksize * BUF_BLOCKS);
 	if (!buf)
 		return ENOMEM;
-	
+
 	for (group = 0; group < fs->group_desc_count; group++) {
-		blk = fs->group_desc[(unsigned)group].bg_inode_table;
+		blk = ext2fs_inode_table_loc(fs, (unsigned)group);
 		if (!blk) {
 			retval = EXT2_ET_MISSING_INODE_TABLE;
 			goto errout;
@@ -164,10 +170,10 @@ errcode_t ext2fs_image_inode_read(ext2_filsys fs, int fd,
 				retval = EXT2_ET_SHORT_READ;
 				goto errout;
 			}
-			retval = io_channel_write_blk(fs->io, blk, c, buf);
+			retval = io_channel_write_blk64(fs->io, blk, c, buf);
 			if (retval)
 				goto errout;
-			
+
 			blk += c;
 			left -= c;
 		}
@@ -182,7 +188,7 @@ errout:
 /*
  * Write out superblock and group descriptors
  */
-errcode_t ext2fs_image_super_write(ext2_filsys fs, int fd, 
+errcode_t ext2fs_image_super_write(ext2_filsys fs, int fd,
 				   int flags EXT2FS_ATTR((unused)))
 {
 	char		*buf, *cp;
@@ -221,7 +227,7 @@ errcode_t ext2fs_image_super_write(ext2_filsys fs, int fd,
 		retval = EXT2_ET_SHORT_WRITE;
 		goto errout;
 	}
-	
+
 	retval = 0;
 
 errout:
@@ -232,7 +238,7 @@ errout:
 /*
  * Read the superblock and group descriptors and overwrite them.
  */
-errcode_t ext2fs_image_super_read(ext2_filsys fs, int fd, 
+errcode_t ext2fs_image_super_read(ext2_filsys fs, int fd,
 				  int flags EXT2FS_ATTR((unused)))
 {
 	char		*buf;
@@ -278,7 +284,7 @@ errout:
 errcode_t ext2fs_image_bitmap_write(ext2_filsys fs, int fd, int flags)
 {
 	ext2fs_generic_bitmap	bmap;
-	errcode_t		err, retval;
+	errcode_t		retval;
 	ssize_t			actual;
 	__u32			itr, cnt, size;
 	int			c, total_size;
@@ -291,7 +297,6 @@ errcode_t ext2fs_image_bitmap_write(ext2_filsys fs, int fd, int flags)
 				return retval;
 		}
 		bmap = fs->inode_map;
-		err = EXT2_ET_MAGIC_INODE_BITMAP;
 		itr = 1;
 		cnt = EXT2_INODES_PER_GROUP(fs->super) * fs->group_desc_count;
 		size = (EXT2_INODES_PER_GROUP(fs->super) / 8);
@@ -302,7 +307,6 @@ errcode_t ext2fs_image_bitmap_write(ext2_filsys fs, int fd, int flags)
 				return retval;
 		}
 		bmap = fs->block_map;
-		err = EXT2_ET_MAGIC_BLOCK_BITMAP;
 		itr = fs->super->s_first_data_block;
 		cnt = EXT2_BLOCKS_PER_GROUP(fs->super) * fs->group_desc_count;
 		size = EXT2_BLOCKS_PER_GROUP(fs->super) / 8;
@@ -314,8 +318,8 @@ errcode_t ext2fs_image_bitmap_write(ext2_filsys fs, int fd, int flags)
 		if (size > (cnt >> 3))
 			size = (cnt >> 3);
 
-		retval = ext2fs_get_generic_bitmap_range(bmap, 
-				 err, itr, size << 3, buf);
+		retval = ext2fs_get_generic_bmap_range(bmap, itr,
+						       size << 3, buf);
 		if (retval)
 			return retval;
 
@@ -324,7 +328,7 @@ errcode_t ext2fs_image_bitmap_write(ext2_filsys fs, int fd, int flags)
 			return errno;
 		if (actual != (int) size)
 			return EXT2_ET_SHORT_READ;
-		
+
 		itr += size << 3;
 		cnt -= size << 3;
 	}
@@ -355,7 +359,7 @@ errcode_t ext2fs_image_bitmap_write(ext2_filsys fs, int fd, int flags)
 errcode_t ext2fs_image_bitmap_read(ext2_filsys fs, int fd, int flags)
 {
 	ext2fs_generic_bitmap	bmap;
-	errcode_t		err, retval;
+	errcode_t		retval;
 	__u32			itr, cnt;
 	char			buf[1024];
 	unsigned int		size;
@@ -368,7 +372,6 @@ errcode_t ext2fs_image_bitmap_read(ext2_filsys fs, int fd, int flags)
 				return retval;
 		}
 		bmap = fs->inode_map;
-		err = EXT2_ET_MAGIC_INODE_BITMAP;
 		itr = 1;
 		cnt = EXT2_INODES_PER_GROUP(fs->super) * fs->group_desc_count;
 		size = (EXT2_INODES_PER_GROUP(fs->super) / 8);
@@ -379,7 +382,6 @@ errcode_t ext2fs_image_bitmap_read(ext2_filsys fs, int fd, int flags)
 				return retval;
 		}
 		bmap = fs->block_map;
-		err = EXT2_ET_MAGIC_BLOCK_BITMAP;
 		itr = fs->super->s_first_data_block;
 		cnt = EXT2_BLOCKS_PER_GROUP(fs->super) * fs->group_desc_count;
 		size = EXT2_BLOCKS_PER_GROUP(fs->super) / 8;
@@ -396,8 +398,8 @@ errcode_t ext2fs_image_bitmap_read(ext2_filsys fs, int fd, int flags)
 		if (actual != (int) size)
 			return EXT2_ET_SHORT_READ;
 
-		retval = ext2fs_set_generic_bitmap_range(bmap, 
-				 err, itr, size << 3, buf);
+		retval = ext2fs_set_generic_bmap_range(bmap, itr,
+						       size << 3, buf);
 		if (retval)
 			return retval;
 
