@@ -38,6 +38,8 @@
  */
 #define _SVID_SOURCE
 
+#include "config.h"
+
 #ifdef _WIN32
 #define _WIN32_WINNT 0x0500
 #include <windows.h>
@@ -86,6 +88,9 @@
 #if defined(__linux__) && defined(HAVE_SYS_SYSCALL_H)
 #include <sys/syscall.h>
 #endif
+#ifdef HAVE_SYS_RESOURCE_H
+#include <sys/resource.h>
+#endif
 
 #include "uuidP.h"
 #include "uuidd.h"
@@ -171,8 +176,7 @@ static void get_random_bytes(void *buf, int nbytes)
 {
 	int i, n = nbytes, fd = get_random_fd();
 	int lose_counter = 0;
-	unsigned char *cp = (unsigned char *) buf;
-	unsigned short tmp_seed[3];
+	unsigned char *cp = buf;
 
 	if (fd >= 0) {
 		while (n > 0) {
@@ -195,12 +199,16 @@ static void get_random_bytes(void *buf, int nbytes)
 	for (cp = buf, i = 0; i < nbytes; i++)
 		*cp++ ^= (rand() >> 7) & 0xFF;
 #ifdef DO_JRAND_MIX
-	memcpy(tmp_seed, jrand_seed, sizeof(tmp_seed));
-	jrand_seed[2] = jrand_seed[2] ^ syscall(__NR_gettid);
-	for (cp = buf, i = 0; i < nbytes; i++)
-		*cp++ ^= (jrand48(tmp_seed) >> 7) & 0xFF;
-	memcpy(jrand_seed, tmp_seed,
-	       sizeof(jrand_seed)-sizeof(unsigned short));
+	{
+		unsigned short tmp_seed[3];
+
+		memcpy(tmp_seed, jrand_seed, sizeof(tmp_seed));
+		jrand_seed[2] = jrand_seed[2] ^ syscall(__NR_gettid);
+		for (cp = buf, i = 0; i < nbytes; i++)
+			*cp++ ^= (jrand48(tmp_seed) >> 7) & 0xFF;
+		memcpy(jrand_seed, tmp_seed,
+		       sizeof(jrand_seed) - sizeof(unsigned short));
+	}
 #endif
 
 	return;
@@ -311,6 +319,7 @@ static int get_clock(uint32_t *clock_high, uint32_t *clock_low,
 	struct flock			fl;
 	uint64_t			clock_reg;
 	mode_t				save_umask;
+	int				len;
 
 	if (state_fd == -2) {
 		save_umask = umask(0);
@@ -392,10 +401,15 @@ try_again:
 
 	if (state_fd > 0) {
 		rewind(state_f);
-		ftruncate(state_fd, 0);
-		fprintf(state_f, "clock: %04x tv: %lu %lu adj: %d\n",
-			clock_seq, last.tv_sec, last.tv_usec, adjustment);
+		len = fprintf(state_f,
+			      "clock: %04x tv: %016lu %08lu adj: %08d\n",
+			      clock_seq, last.tv_sec, (long)last.tv_usec,
+			      adjustment);
 		fflush(state_f);
+		if (ftruncate(state_fd, len) < 0) {
+			fprintf(state_f, "                   \n");
+			fflush(state_f);
+		}
 		rewind(state_f);
 		fl.l_type = F_UNLCK;
 		fcntl(state_fd, F_SETLK, &fl);
@@ -411,15 +425,19 @@ static ssize_t read_all(int fd, char *buf, size_t count)
 {
 	ssize_t ret;
 	ssize_t c = 0;
+	int tries = 0;
 
 	memset(buf, 0, count);
 	while (count > 0) {
 		ret = read(fd, buf, count);
-		if (ret < 0) {
-			if ((errno == EAGAIN) || (errno == EINTR))
+		if (ret <= 0) {
+			if ((errno == EAGAIN || errno == EINTR || ret == 0) &&
+			    (tries++ < 5))
 				continue;
-			return -1;
+			return c ? c : -1;
 		}
+		if (ret > 0)
+			tries = 0;
 		count -= ret;
 		buf += ret;
 		c += ret;
@@ -427,6 +445,33 @@ static ssize_t read_all(int fd, char *buf, size_t count)
 	return c;
 }
 
+/*
+ * Close all file descriptors
+ */
+static void close_all_fds(void)
+{
+	int i, max;
+
+#if defined(HAVE_SYSCONF) && defined(_SC_OPEN_MAX)
+	max = sysconf(_SC_OPEN_MAX);
+#elif defined(HAVE_GETDTABLESIZE)
+	max = getdtablesize();
+#elif defined(HAVE_GETRLIMIT) && defined(RLIMIT_NOFILE)
+	struct rlimit rl;
+
+	getrlimit(RLIMIT_NOFILE, &rl);
+	max = rl.rlim_cur;
+#else
+	max = OPEN_MAX;
+#endif
+
+	for (i=0; i < max; i++) {
+		close(i);
+		if (i <= 2)
+			open("/dev/null", O_RDWR);
+	}
+}
+
 
 /*
  * Try using the uuidd daemon to generate the UUID
@@ -442,6 +487,7 @@ static int get_uuid_via_daemon(int op, uuid_t out, int *num)
 	ssize_t ret;
 	int32_t reply_len = 0, expected = 16;
 	struct sockaddr_un srv_addr;
+	struct stat st;
 	pid_t pid;
 	static const char *uuidd_path = UUIDD_PATH;
 	static int access_ret = -2;
@@ -457,8 +503,13 @@ static int get_uuid_via_daemon(int op, uuid_t out, int *num)
 		    sizeof(struct sockaddr_un)) < 0) {
 		if (access_ret == -2)
 			access_ret = access(uuidd_path, X_OK);
+		if (access_ret == 0)
+			access_ret = stat(uuidd_path, &st);
+		if (access_ret == 0 && (st.st_mode & (S_ISUID | S_ISGID)) == 0)
+			access_ret = access(UUIDD_DIR, W_OK);
 		if (access_ret == 0 && start_attempts++ < 5) {
 			if ((pid = fork()) == 0) {
+				close_all_fds();
 				execl(uuidd_path, "uuidd", "-qT", "300",
 				      (char *) NULL);
 				exit(1);
