@@ -11,6 +11,13 @@
 #define const
 #endif
 
+#ifdef __GNUC__
+#define FSCK_ATTR(x) __attribute__(x)
+#else
+#define FSCK_ATTR(x)
+#endif
+
+
 #ifndef DEFAULT_FSTYPE
 #define DEFAULT_FSTYPE	"ext2"
 #endif
@@ -61,3 +68,6 @@ struct fsck_instance {
 
 extern char *base_device(const char *device);
 extern const char *identify_fs(const char *fs_name, const char *fs_types);
+
+/* ismounted.h */
+extern int is_mounted(const char *file);
